
/*   ----------------------------------------------------
     NOTES
     5-19-09: changed database format to a 2-D array
     5-19-09: added more tables to the database
     6-2-09: started making notes for the MULTI version

     TODO: 
  2. delete duplicate entries as we display them
     a. NO - delete duplicate entries from the database before inclusion
  10. Add a "show these on a map" link
      a. requires a server. how do we communicate? post?
  14. Almost 300 entries are not geocoded. The city names are misspelled.
      Zip codes have the wrong number of digits. Street names misspelled.
  16. MULTI project: allow searching through multiple databases.

     Partly FIXED:
  1. geo.html has a bug where all "+" do not get replaced
     a. Fixed on the Powerbook; merge fix back onto Tower 
  7. make the code universal so we can use it for NPS etc.

     DONE:
  11. This database does not include Canada! (some do)
  12. The static geocoder does not include Canada!
  13. Encode health_data etc. as global variables
  15. Store strings in the database: title, "stores found within", 
      "show more stores", 
     ---------------------------------------------------- */

//-------------------------------------------------------
// global constants and variables
//------------------------------------------------------- 
// TODO: MULTI requires that there be NO global variables.
//------------------------------------------------------- 
var reqlat, reqlong, reqmax, reqmiles, reqcity, reqstate;
var notfound;

// TODO: maybe we should use a "flipped" array instead
var fld_source, fld_link, fld_name, fld_address, fld_city, fld_state;
var fld_near, fld_zipcode, fld_phone, fld_lat, fld_long;

//-------------------------------------------------------
// main code
//-------------------------------------------------------
// TODO: what the heck?
// window.onload = function () {
// window.onload = main;
// main ();
function main () {
  get_params (); 
  form_setup (self, geocoder_url, reqcity, reqstate);

  // TODO: MULTI version requires that this field_setup be done
  // for each database individually
  field_setup (mydb.fields);
  if (reqlat && reqlong && !notfound) {

    // TODO: MULTI requires that this be done for each database.
    // PROBLEM: there may be conflicts between the databases.

    if (mydb.css)
      import_styles (mydb.css);
    show_request (reqcity, reqstate, reqlat, reqlong, reqmax, reqmiles);

    // TODO: MULTI version requires that we can find matches
    // in one particular database.  
    var matches = find_matches (reqlat, reqlong);

    get_distances (matches, reqlat, reqlong);
    matches.sort (numeric_sort);
    var num_matches = show_results (matches, reqmax, reqmiles);
    show_more_links (reqmax, reqmiles, num_matches);

    // TODO: MULTI ??
    // I think show_legend requires a different kind of icon
    // that indicates which of the MULTI databases this record
    // came from. 
    show_legend (mydb);
  }
  else if (reqlat && reqlong) {
    show_request_notfound(reqcity, reqstate, reqlat, reqlong, reqmax, reqmiles);
  }
} 

//-------------------------------------------------------
// numeric_sort: sort numeric instead of alpha (callback)
//-------------------------------------------------------
function numeric_sort (a, b)
{
  return (a[1] - b[1]);
}

//-------------------------------------------------------
// form_setup: add characteristics to the static form
//-------------------------------------------------------
function form_setup (self, geocoder_url, city, state)
{
  document.getElementById ("return").value = self;
  document.getElementById ("form").action = geocoder_url;
  if (reqcity)
    document.getElementById ("city").value = unescape (reqcity);
  document.getElementById ("state").value = reqstate;

  // TODO: MULTI requires that the strings come from a META db
  // set up title and strings from the database
  document.title = mydb.strings["title"];
  muSetInnerText (
    document.getElementsByTagName ("h1")[0], mydb.strings["title"]);
  var i;
  for (i=0; i<document.getElementsByClassName ("items").length; i++) {
    muSetInnerText (document.getElementsByClassName ("items")[i],
      mydb.strings["items"]);
  }
} 

//-------------------------------------------------------
// field_setup (a)
//-------------------------------------------------------
// TODO: MULTI version requires that this information
// be unique for each database to be searched.
//-------------------------------------------------------
function field_setup (a)
{
  // TODO: maybe we should use a "flipped" array instead
  var i,j;
  for (i=0; i<a.length; i++) {
    if (a[i] == "source")    fld_source = i;
    if (a[i] == "link")      fld_link = i;
    if (a[i] == "name")      fld_name = i;
    if (a[i] == "near")      fld_near = i;
    if (a[i] == "address")   fld_address = i;
    if (a[i] == "city")      fld_city = i;
    if (a[i] == "state")     fld_state = i;
    if (a[i] == "zipcode")   fld_zipcode = i;
    if (a[i] == "phone")     fld_phone = i;
    if (a[i] == "latitude")  fld_lat = i;
    if (a[i] == "longitude") fld_long = i;
  }
}

//-------------------------------------------------------
// import_styles (a)
//-------------------------------------------------------
function import_styles (a)
{
  var i;
  for (i=0; i<a.length; i++) {
    if (a[i] != "")
      //add a new rule at the start
      document.styleSheets[0].insertRule(a[i],1); /* changed 0 to 1 on 7-21-14 */
  }
  // alert (document.styleSheets[0].rules[0]);
}

//-------------------------------------------------------
// show_more_links
//-------------------------------------------------------
function show_more_links (max, miles, num)
{
  var url = self + "?latitude=" + reqlat + "&longitude=" + reqlong +
    "&city=" + reqcity + "&state=" + reqstate;
  if (num < max) // need more miles
    url = url + "&max=" + max + "&miles=" + (2*miles);
  else // need more matches
    url = url + "&max=" + (2*max) + "&miles=" + miles;

  // TODO: MULTI requires that the strings come from a META db
  document.getElementById ("showmore").innerHTML =
    num + " " + mydb.strings["items"] + " found within " + miles + 
    " miles. " + "<a href=\"" + url + "\">" + 
    "Show more " + mydb.strings["items"] + " ..." + "<\/a>\n";
}

//-------------------------------------------------------
// show_request: show the requested parameters including
// the calculated latitude and longitude
//-------------------------------------------------------
function show_request (city, state, lat, long, max, miles)
{
  var latlong_text = 
    "Latitude " + lat + "; Longitude " + long;
  document.getElementById ("latlong").innerHTML = latlong_text;
}

//-------------------------------------------------------
// show_request_notfound: show an error message.
//-------------------------------------------------------
function show_request_notfound (city, state, lat, long, max, miles)
{
  var latlong_text = 
    "<p><center>Sorry, <b>" + unescape(city) + ", " + state + 
    "</b> not found. Try again.<\/center>";
  document.getElementById ("latlong").innerHTML = latlong_text;
}

//-------------------------------------------------------
// find_lat: find the record nearest this latitude
//-------------------------------------------------------
function find_lat (a, lats, lat)
{
  return (find_field (a, lats, lat, fld_lat));
}

//-------------------------------------------------------
// find_long: find the record nearest this longitude
//-------------------------------------------------------
function find_long (a, longs, long)
{
  return (find_field (a, longs, long, fld_long));
}

//-------------------------------------------------------
// find_field: find the record nearest this longitude OR latitude
//-------------------------------------------------------
function find_field (a, order, spec, field, first, last)
{
  // Binary Search through ordered latitudes or longitudes
  // TODO: which one we return depends on whether first or last
  // If first, we should return the second of the pair found.
  // If last, we should return the first of the pair found.
  if (first === undefined)
    first = 0;
  if (last === undefined)
    last = order.length - 2;
  if (last-first<2) {
    return first;
  }

  var mid = Math.floor((first+last)/2);
  var latlong_1b = a[order[mid]][field];
  var latlong_2b = a[order[mid+1]][field];

  if (latlong_1b <= spec && latlong_2b >= spec) {
    return (mid);
  }
  else if (latlong_1b <= spec && latlong_2b <= spec) {
    return (find_field (a, order, spec, field, mid, last));
  }
  else
    return (find_field (a, order, spec, field, first, mid)); {
  }
}

//-------------------------------------------------------
// function array_intersect: compute the intersection of two arrays of numbers
// NOTE: sorts each array in place, and returns a sorted array
// NOTE: treats each array as a set rather than an ordered list.
//-------------------------------------------------------
function array_intersect (a, b)
{
  // document.write ("pre-sort<br>");
  // document.write ("A<br> " + a + "<p>B<br> " + b + "<p>");
  // 1. sort each array
  a.sort (numsortfunc);
  b.sort (numsortfunc);
  // document.write ("post-sort<br>");
  // document.write ("A<br> " + a + "<p>B<br> " + b + "<p>");

  // 2. merge sort including only elements that are in both arrays
  var na = a.length;
  var nb = b.length;
  var c = new Array ();

  i=j=0;
  while (i<na && j<nb) {
    // document.write ("INDEX-A " + i + " INDEX-B " + j
    if (a[i] == b[j]) {
      c.push (a[i]);
      i++;
      j++;
    }
    else if (a[i] < b[j]) {
      i++;
    }
    else if (a[i] > b[j]) {
      j++;
    } 
    else {
      break;
    }
  } 
  return (c);
}

// ----------- util function (callback) ----------------------------
function numsortfunc (a, b) { return a - b; }

//-------------------------------------------------------
// find_matches: find records near this lat and long
//-------------------------------------------------------
// TODO: MULTI version requires that we can find matches
// in one particular database.
//-------------------------------------------------------
function find_matches (lat, long)
{
  // for testing purposes
  // var brute = find_matches_brute_force (lat, long);
  // alert (brute);
  // return (brute);

  // TODO: for longitudes near 180 degress, the +2 and -2 math might fail.
  // We need to check for abs(long+2) > 180, and change to 
  // or result-360 or result+360
  // e.g. 181 -> -179 ...... result-360 if result > 0
  // and  -181 -> 179 ...... result+360 if result < 0
  // or -1 * sign(result) * 360 + result => new result
  // where sign(result) = (abs(result))/result
  // NOTE the only real problem country (for dance clubs, anyway)
  // would be New Zealand, and those longitudes extend only to 176.99

  // 1. find all results within 2 degrees of lat
  var first_lat = find_lat (mydb.data, mydb.data_lat, lat-2);
  var last_lat = find_lat (mydb.data, mydb.data_lat, lat-0+2);
  var lat_a = mydb.data_lat.slice (first_lat, last_lat);
  // document.write ("LATS " + first_lat + " " + last_lat + "<br>" + lat_a + "<p>");

  // 2. find all results within 2 degrees of long
  var first_long = find_long (mydb.data, mydb.data_long, long-2);
  var last_long = find_long (mydb.data, mydb.data_long, long-0+2);
  var long_a = mydb.data_long.slice (first_long, last_long);
  // document.write ("LONGS " + first_long + " " + last_long + "<br>" + long_a + "<p>");

  // 3. array_intersect
  var matches_z = array_intersect (lat_a, long_a);
  // document.write (matches_z);

  var i;
  var rv = new Array ();
  var temp = new Array ();
  for (i=0; i<matches_z.length; i++) {
    temp = [];
    temp.push (matches_z[i]);
    temp.push (-1); // placeholder for distance
    rv.push (temp);
  }
  return (rv);
}

//-------------------------------------------------------
function find_matches_brute_force (lat, long)
//-------------------------------------------------------
{
  /* 
   * This search method does not make use of the indexes. SLOW.
   */
  var i;
  var rv = new Array();
  var temp = new Array ();
  for (i=0; i<mydb.data.length; i++) {
    var test_lat = mydb.data[i][fld_lat];
    var test_long = mydb.data[i][fld_long];
    if ((Math.abs (test_lat - lat) <= 2) &&
        (Math.abs (test_long - long) <= 2))
    {
      temp = [];
      temp.push (i);
      temp.push (-1);
      rv.push (temp);
    }
  }
  return (rv);
}

//-------------------------------------------------------
// get_distances: for each entry in the array a, find the distance 
// from the location (lat, long). Add this distance to the array "a" 
// as the 2nd field in each data record (which should be blank beforehand).
//-------------------------------------------------------
// TODO: MULTI requires that we be able to do this for multiple mydb's.
//-------------------------------------------------------
function get_distances (a, lat, long)
{
  var i;
  var dist;

  for (i=0; i<a.length; i++) {
    var record = mydb.data[a[i][0]];
    dist = distance_yxyx (record[fld_lat], record[fld_long], lat, long);
    a[i][1] = dist;
  }
}

//-------------------------------------------------------
// printf (format, data): replace %s in the format string with data
// TODO: data might be an array of substitutions
//-------------------------------------------------------
function printf (format, data)
{
  return (format.replace ("%s", data));
}

//-------------------------------------------------------
// make_database_link: use this entry to create a link
//   to the original scraped data page
//-------------------------------------------------------
function make_database_link (record)
{
  // TODO: modified 5-21-09 for the churches database.
  // Might have some bugs now.
  // Modified 5-26-09 for the GF restaurants database.
  // was "if (fld_link)" which doesn't work for link in the 0 field.

  var url="";
  var subber;
  var prefix = get_url_prefix (record[fld_source]);
  if (fld_link != undefined)
    subber = record[fld_link];
  else
    subber = "";
  if (subber.substr(0,4) == "http")
    url = subber;
  else if (prefix.search ("%s") && (subber != "-")) 
    url = printf (prefix, subber); 
  else if (prefix != "" && prefix != "%s")
    url = prefix;
 
  var record_name;
  if (fld_name == undefined && mydb.strings.defaultname)
    record_name = mydb.strings.defaultname;
  else
    record_name = record[fld_name]; 

  var link;
  if (url != "")
    link = 
      "<b><a href=\"" + url + "\">" + record_name + "<\/a><\/b><br>\n";
  else
    link = "<b>" + record_name + "<\/b><br>\n";
  return (link);
}

//-------------------------------------------------------
// make_google_map_link: use this address to create a
//   Google Map query
//-------------------------------------------------------
function make_google_map_link (record)
{
  //
  // Modified 7-24-14 to get rid of problematic addresses like
  // "2 miles NE of Livingston, CA"
  // and replace them with the Latitude and Longitude, which we also already have.
  // 
  if (mydb.prefs && mydb.prefs.latlong == true)
    return (make_latlong_link (record));

  var url = "http://maps.google.com/?hl=en&q=";
  var addr = "";
  var near = "";                             // 7-24-14
  if (record[fld_near])
    // addr = addr + record[fld_near] + " "; // 7-24-14
    near = record[fld_near] + " ";           // 7-24-14
  if (record[fld_address])
    addr = addr + record[fld_address] + ", ";
  addr = addr + record[fld_city] + ", ";
  addr = addr + record[fld_state];
  if (record[fld_near])                                   // 7-24-14
    url = url + record[fld_lat] + "," + record[fld_long]; // 7-24-14
  else if (!record[fld_address])
    url = url + record[fld_lat] + "," + record[fld_long]; // 7-24-14
  else                                                    // 7-24-14
    url = url + addr;
  var link = "<a href=\"" + url + "\">" + near + addr + "<\/a>"; // 7-24-14
  return (link);
}

function make_latlong_link (record)
{
  var url = "http://maps.google.com/?hl=en&z=10&q=";
  url = url + record[fld_lat] + ", ";
  url = url + record[fld_long] + " (";
  url = url + record[fld_name] + ")";
  var addr = "";
  if (record[fld_near])
    addr = addr + record[fld_near] + " ";
  if (record[fld_address])
    addr = addr + record[fld_address] + ", ";
  addr = addr + record[fld_city] + ", ";
  addr = addr + record[fld_state];
  var link = "<a href=\"" + url + "\">" + addr + "<\/a>";
  return (link);
}

//-------------------------------------------------------
// show_one_result
//-------------------------------------------------------
// TODO: MULTI requires that we allow for multiple mydb's.
//-------------------------------------------------------
function show_one_result (record, distance)
{
  if (mydb.show_one_result)
    return (mydb.show_one_result (record, distance));

  var contents = "";

  // icon (reverse color lettering)
  if (mydb.icons)
    contents += (get_icon (record[fld_source]) + "\n");

  // Name of store (with link to the original source web page)
  contents += (make_database_link (record));

  // Street address (with link to Google Map)
  contents += (make_google_map_link (record));

  // zip code
  if (fld_zipcode && record[fld_zipcode] != "")
    contents += (", " + record[fld_zipcode]);
  contents += ("<br>\n");

  // Phone number
  if (fld_phone && record[fld_phone] != "")
    contents += (record[fld_phone] + "<br>\n");

  if (mydb.show_one_result_extras)
    contents += mydb.show_one_result_extras (record);
    
  // distance and direction
  dir = direction_yxyx (reqlat, reqlong, record[fld_lat], record[fld_long]);
  contents += (
    // "Distance: " + roundN (matches[i][1],2) + " miles " + dir + "<br>\n");
    "Distance: " + roundN (distance,2) + " miles " + dir + "<br>\n");
  // contents += ("<p>\n"); 
  contents = "<p>" + contents + "<\/p>\n";
  return (contents);
}

//-------------------------------------------------------
// show_results: display the given matches
//-------------------------------------------------------
// TODO: MULTI requires that we handle multiple mydb's.
//-------------------------------------------------------
function show_results (matches, max, miles) 
{ 
  // ----------------------------------------------------
  // TODO: the display mechanism ought to be a template
  //   within the data file
  // TODO: there could be multiple ways of displaying the list
  // TODO: compare this record against the last one and
  //   don't display duplicate entries
  // ----------------------------------------------------
  // display the ones within this number and distance 
  var i, j;
  var data;
  var latlong, lat, long;
  var num_matches=0;
  var contents = "";
  for (i=0; i<matches.length && i<max; i++) {
    if (matches[i][1] > miles)
      break;

    // get the record we are going to print
    var record = mydb.data[matches[i][0]];
    contents += show_one_result (record, matches[i][1]); 
    num_matches++;
  }
  // contents = "<table><tr><td>\n" + contents + "\n<\/td><\/tr><\/table>\n";
  document.getElementById ("results").innerHTML = contents;
  return (num_matches);
}

// ------------------------------------------------------------
// get_url_prefix
// ------------------------------------------------------------
// TODO: MULTI requires that we allow for multiple mydb's.
// ------------------------------------------------------------
function get_url_prefix (c)
{
  if (mydb.prefixes)
    return (mydb.prefixes[c]);
  else if (mydb.strings.baseurl)
    return (mydb.strings.baseurl); 
  else
    return ("");
}

// ------------------------------------------------------------
// get_icon
// ------------------------------------------------------------
// TODO: MULTI requires that we allow for multiple mydb's.
// Also, we might want to have a META-db that displays a 
// different kind of icon that corresponds to the database 
// this record came from.
// ------------------------------------------------------------
function get_icon (c)
{
  return (mydb.icons[c]);
}

// ------------------------------------------------------------
function get_params ()
// ------------------------------------------------------------
{
  // get the latitude=, longitude=, max=, and miles= params from the URL
  // NOTE: need also maybe to get city= and state=

  var latsearch = /latitude=([-\.0-9]*)/
  var longsearch = /longitude=([-\.0-9]*)/
  var maxsearch = /max=([0-9]*)/
  var milesearch = /miles=([0-9]*)/
  var citysearch = /city=([A-Za-z0-9%+\.-]*)/
  var statesearch = /state=([A-Z]*)/

  // assign to global variables
  reqlat = location.search.match (latsearch);
  reqlong = location.search.match (longsearch);

  reqmax = location.search.match (maxsearch);
  reqmiles = location.search.match (milesearch);
  if (! reqmax) reqmax = 10; else reqmax = reqmax[1];
  if (! reqmiles) reqmiles = 30; else reqmiles = reqmiles[1];

  reqcity = location.search.match (citysearch);
  reqstate = location.search.match (statesearch);

  if (reqlat) reqlat = reqlat[1];
  if (reqlong) reqlong = reqlong[1];
  if (reqcity) reqcity = reqcity[1];
  if (reqstate) reqstate = reqstate[1];

  notfound = false;
  if (location.search.search ("notfound=1") > -1)
    notfound = true;
}

//---------------------------------------------------------------------
function distance_yxyx (y1, x1, y2, x2)
//---------------------------------------------------------------------
{
  // ---
  // compute the distance in miles between these two lat/long coordinates
  // ---

  // http://www.meridianworlddata.com/Distance-Calculation.asp
  a = Math.sin (y1/57.296) * Math.sin (y2/57.296);
  b = Math.cos (y1/57.296) * Math.cos (y2/57.296) * 
      Math.cos ((x2/57.296) - (x1/57.296));
  return  3963.0 * Math.acos(a+b) ;
}

//---------------------------------------------------------------------
function roundN (f, digits)
//---------------------------------------------------------------------
{
  var temp = f * Math.pow (10, digits);
  temp = Math.round (temp);
  temp = temp / Math.pow (10, digits);
  return (temp);
}

//---------------------------------------------------------------------
function direction_yxyx (y1, x1, y2, x2)
//---------------------------------------------------------------------
{
  // ---
  // divide the clock into eight slices and return the direction
  //   from point (x,y) to (x2,y2).
  // The result can be one of N W E S NW SW NE SE
  // ---
  var dy = y2 - y1;
  var dx = x2 - x1;
  var dir_y, dir_x;
  if (dy > 0) dir_y = "N"; else dir_y = "S";
  if (dx > 0) dir_x = "E"; else dir_x = "W";   
  if (dy == 0)
    return (dir_x);
  var ratio = Math.abs (dx/dy);
  if (ratio > 2)
    dir = dir_x;
  else if (ratio < 0.5)
    dir = dir_y;
  else
    dir = dir_y + dir_x;
  return (dir);
} 

//---------------------------------------------------------------------
function muSetInnerText (node, value)
//---------------------------------------------------------------------
{   
  if (node.innerText) node.innerText = value; // Safari
  else if (node.textContent) node.textContent = value; // Firefox
} 
  
//---------------------------------------------------------------------
function muGetInnerText (node)
//---------------------------------------------------------------------
{ 
  if (node.innerText) return (node.innerText); // Safari
  else if (node.textContent) return (node.textContent); // Firefox
}


//---------------------------------------------------------------------
function show_legend (mydb)
//---------------------------------------------------------------------
{
  if (!mydb.legend) 
    return;

  var i;
  var contents = "";
  contents += "<p><b>Legend</b><p>\n";
  for (i=0; i<mydb.legend.length; i++) {
    if (mydb.icons)
      contents += (get_icon (mydb.legend[i][0]) + "\n");
    if (mydb.legend[i][2]) {
      contents += 
       "<a href=\"" + mydb.legend[i][2] + "\">" + mydb.legend[i][1] +
       "<\/a><br>\n";
    }
    else {
      contents += mydb.legend[i][1] + "<br>\n";
    }
  }
  // contents = "<table><tr><td>\n" + contents + "\n<\/td><\/tr><\/table>\n";

  document.getElementById ("legend").innerHTML = contents;
}

